% Companion software to the paper
% "Computing the Reciprocal of a $\phi$ Function 
% by Rational Approximation",
% by P. Boito, Y. Eidelman and L. Gemignani.
%
% Compares scalar approximations to the psi_1 function.
% Chebfun and EXPINT toolboxes are needed;
% you will need to download them and add paths as necessary.
%
% Last modified Sept 19, 2019.

clear all
close all

% addpath /home/paola/Documents/Documents/MATLAB/luca_phi/expint/common/
% addpath /home/paola/Documents/Documents/MATLAB/luca_krylov/chebfun-master/

m = 6;  % degree of diagonal Padé 
np=3;   % degrees of mixed approx
ns=20;

f = @(z)z./(exp(z)-1); % psi_1
g = @(z)(exp(z)-1)./z; % phi_1

% Define grid
ngrid=80;
xgrid = linspace(-20,20,ngrid);
[x,y] = meshgrid(xgrid,xgrid);
z = x + 1i*y;

% Classical Padé from Chebfun
figure(1)
[r,a,b] = padeapprox(f,m,m,0);
contourf(x,y,max(-16,log10(abs(f(z) - r(z)))),[-16:2:0])
caxis([-16,2]);
colorbar
title('Padé (no scaling)')
set(gca,'FontSize',18)  % set large font for use in paper

% Padé with scaling from Expint
figure(2)
rpade=full(diag(phipade(z(:),1,m)));
rrpade=reshape(rpade,80,80);
contourf(x,y,max(-16,log10(abs(f(z) - 1./rrpade))),[-16:2:0])
caxis([-16,2]);
colorbar
title('Padé with scaling')
set(gca,'FontSize',18)  % set large font for use in paper

% Our approximation without scaling
figure(3)
w=our_rat_eval(np,ns,z);  
contourf(x,y,max(-16,log10(abs(f(z) - w))),[-16:2:0])
caxis([-16,2]);
colorbar
title('Mixed approximation')
set(gca,'FontSize',18)  % set large font for use in paper

% Our approximation with scaling
figure(4)
ws=our_rat_eval_ws(np,ns,z);  
contourf(x,y,max(-16,log10(abs(f(z) - ws))),[-16:2:0])
caxis([-16,2]);
colorbar
title('Mixed approximation with scaling')
set(gca,'FontSize',18)  % set large font for use in paper

% Carathéodory-Fejér as in Trefethen & Schmelzer
figure(5)
[zi,ci,r_inf] = cf_rat_phi1(2*m);   % should this be the correct degree?
zz=z(:).';
cf=r_inf+sum(ci./(zz-zi));
cfapprox=reshape(cf.',80,80);
contourf(x,y,max(-16,log10(abs(f(z) - 1./cfapprox))),[-16:2:0])
caxis([-16,2]);
colorbar
title('Carathéodory-Fejér, TS')
set(gca,'FontSize',18)  % set large font for use in paper

% Carathéodory-Fejér as in Chebfun
figure(6)
chebpsi=chebfun(f);
clear cf;
[p,q,rhandle,s]=cf(chebpsi,m,m);
zz=z(:);
cf2=rhandle(zz);
cfapprox2=reshape(cf2,80,80);
contourf(x,y,max(-16,log10(abs(f(z) - cfapprox2))),[-16:2:0])
caxis([-16,2]);
colorbar
title('Carathéodory-Fejér, Chebfun')
set(gca,'FontSize',18)  % set large font for use in paper
