% Implements Thm 18.7 from Yuli's book
% (inversion of an r-qs matrix).
%
% Generator sizes:
% p is Nxr,
% a is rxrxN,
% q is rxN,
% g is Nxr,
% b is rxrxN,
% h is rxN,
% d is Nx1.
%

function [newp,newa,newq,newg,newb,newh,newd]=invert_187kk_function(p,a,q,g,b,h,d)

N=length(d);  % matrix size
r=length(a(:,1,1));
% Warning: no check on consistency of generator sizes!

% **** Check: ****
% A=recover_matrix_k(p,q,a,g,h,b,d);
% ****************


% allocate new generators
newp=zeros(N,r);
newq=zeros(r,N);
newg=zeros(N,r);
newh=zeros(r,N);
newa=zeros(r,r,N);
newb=zeros(r,r,N);
newd=zeros(N,1);

% allocate auxiliary variables
gamma=zeros(N,1);
f=zeros(r,r,N);
lambda=zeros(r,r,N);

% implement formulas
gamma(1)=d(1);
newq(:,1)=q(:,1)/gamma(1);
newg(1,:)=g(1,:)/gamma(1);
f(:,:,1)=newq(:,1)*gamma(1)*newg(1,:);
for k=2:N-1
    gamma(k)=d(k)-p(k,:)*f(:,:,k-1)*h(:,k);
    newq(:,k)=(q(:,k)-a(:,:,k)*f(:,:,k-1)*h(:,k))/gamma(k);
    newg(k,:)=(g(k,:)-p(k,:)*f(:,:,k-1)*b(:,:,k))/gamma(k);
    f(:,:,k)=a(:,:,k)*f(:,:,k-1)*b(:,:,k)+newq(:,k)*gamma(k)*newg(k,:);
end
gamma(N)=d(N)-p(N,:)*f(:,:,N-1)*h(:,N);

lambda(:,:,N)=-h(:,N)*p(N,:)/gamma(N);
newh(:,N)=h(:,N)/gamma(N);
newp(N,:)=-p(N,:)/gamma(N);
newd(N)=1/gamma(N);
for k=N-1:-1:2
    newa(:,:,k)=a(:,:,k)-newq(:,k)*p(k,:);
    newb(:,:,k)=b(:,:,k)-h(:,k)*newg(k,:);
    lambda(:,:,k)=-h(:,k)*p(k,:)/gamma(k)+newb(:,:,k)*lambda(:,:,k+1)*newa(:,:,k);
    newh(:,k)=h(:,k)/gamma(k)+newb(:,:,k)*lambda(:,:,k+1)*newq(:,k);
    newp(k,:)=-p(k,:)/gamma(k)-newg(k,:)*lambda(:,:,k+1)*newa(:,:,k);
    newd(k)=1/gamma(k)-newg(k,:)*lambda(:,:,k+1)*newq(:,k);
end
newd(1)=1/gamma(1)-newg(1,:)*lambda(:,:,2)*newq(:,1);

%%%%%%%%%%%%%%%%
% edit needed
% (mistake in formulas?)
 newh=-newh;
%%%%%%%%%%%%%%%%

% **** Check: ****
% B=recover_matrix_k(newp,newq,newa,newg,newh,newb,newd);
% ****************